//-< JAVAGO.H >------------------------------------------------------*--------*
// JavaGO                     Version 1.02       (c) 1998  GARRET    *     ?  *
// (Java Global Optimizer)                                           *   /\|  *
//                                                                   *  /  \  *
//                          Created:     10-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 11-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*

#ifndef __JAVAGO_H__
#define __JAVAGO_H__

#define VERSION 1.03

typedef unsigned char  byte;
typedef unsigned short word;

#define bool  int 
#define true  1
#define false 0


enum vbm_instruction_code { 
#define JAVA_INSN(code, mnem, len) mnem,
#include "javago.d"
last_insn
};
    

#define MAX_METHOD_PARAMS 256

inline int unpack2(byte* s) { 
    return (s[0] << 8) + s[1]; 
}

inline int unpack4(byte* s) { 
    return (((((s[0] << 8) + s[1]) << 8) + s[2]) << 8) + s[3];
} 

inline int unpack2_le(byte* s) { 
    return (s[1] << 8) + s[0]; 
}

inline int unpack4_le(byte* s) { 
    return (((((s[3] << 8) + s[2]) << 8) + s[1]) << 8) + s[0];
} 

inline byte* pack2(byte* d, int val) { 
    *d++ = byte(val >> 8);
    *d++ = byte(val);
    return d;
}

inline byte* pack4(byte* d, int val) { 
    *d++ = byte(val >> 24);
    *d++ = byte(val >> 16);
    *d++ = byte(val >> 8);
    *d++ = byte(val);
    return d;
}

extern unsigned string_hash_function(byte* p, int len);

class utf_string {
  public:
    int   len;
    byte* data;

    bool operator == (utf_string const& str) const { 
	return len == str.len && memcmp(data, str.data, len) == 0; 
    }

    bool operator != (utf_string const& str) const { 
	return len != str.len || memcmp(data, str.data, len) != 0; 
    }

    bool operator == (const char* str) const { 
	int slen = strlen(str);
	return len == slen && memcmp(data, str, slen) == 0; 
    }

    bool operator != (const char* str) const { 
	int slen = strlen(str);
	return len != slen || memcmp(data, str, slen) != 0; 
    }

    unsigned hash() const { 
	return string_hash_function(data, len);
    }

    utf_string(int length, byte* str) { 
	len = length;
	data = str;
    }
};

enum const_types { 
    c_none,
    c_utf8,
    c_reserver,
    c_integer,
    c_float,
    c_long,
    c_double, 
    c_class,
    c_string,
    c_field_ref,
    c_method_ref,
    c_interface_method_ref,
    c_name_and_type
};

enum operand_types { 
    t_int,
    t_object,
    t_float, 
    t_double,
    t_long
};

enum access_flags { 
    f_static       = 0x0008,
    f_synchronized = 0x0020,
    f_volatile     = 0x0040,
    f_native       = 0x0100,
    f_abstract     = 0x0400
};

class class_file;

class generation_context { 
  public:
    int   max_locals;
    int   max_stack;
    int   constant_pool_count;
    int   mapped_variables;
    byte* exception_ptr;
    byte* constant_ptr;
    byte* code_ptr;
    class_file* this_class;
    int   variable_mapping[MAX_METHOD_PARAMS];

    byte *exception_buffer, *exception_buffer_end;
    byte *constant_buffer, *constant_buffer_end;
    byte code_buffer[0x10000];

    byte* extend_exception_buffer(int size);
    byte* extend_constant_buffer(int size);

    void reset_for_new_method();
    void reset_for_new_class(class_file* cls);

    generation_context();
};    


class constant { 
  public: 
    int tag;
    constant(byte* p) { tag = *p; }

    virtual int length() = 0;

    virtual int copy(generation_context& ctx, constant** pool, int* mapping)=0;

    static int  relocate(int index, generation_context& ctx, constant** pool,
			 int* mapping)
    {
	int new_index = mapping[index];
	if (new_index == 0) { 
	    new_index = pool[index]->copy(ctx, pool, mapping); 
	    mapping[index] = new_index;
	} 
	return new_index;
    }
};

class const_utf8 : public constant {
  public:
    utf_string str;
    const_utf8(byte* p) : constant(p), str(unpack2(p+1), p+3) {}

    virtual int length() { return 3 + str.len; }

    virtual int copy(generation_context& ctx, constant**, int*) { 
	byte* dst = ctx.extend_constant_buffer(3 + str.len);
	*dst++ = tag;
	dst = pack2(dst, str.len);
	memcpy(dst, str.data, str.len);
	return ctx.constant_pool_count++;
    }
};

class const_literal : public constant {
  public:
    byte* ptr;
    const_literal(byte* p) : constant(p) {
	ptr = p+1;
    }
    virtual int length() { return 5; }

    virtual int copy(generation_context& ctx, constant**, int*) { 
	byte* dst = ctx.extend_constant_buffer(5);
	dst[0] = tag;
	dst[1] = ptr[0];
	dst[2] = ptr[1];
	dst[3] = ptr[2];
	dst[4] = ptr[3];
	return ctx.constant_pool_count++;
    }
};

class const_long_literal : public constant {
  public:
    byte* ptr;
    const_long_literal(byte* p) : constant(p) {
	ptr = p+1;
    }
    virtual int length() { return 9; }

    virtual int copy(generation_context& ctx, constant**, int*) { 
	byte* dst = ctx.extend_constant_buffer(9);
	*dst++ = tag;
	memcpy(dst, ptr, 8);
	int index = ctx.constant_pool_count;
	ctx.constant_pool_count += 2;
	return index;
    }
};

class const_string : public constant { 
  public: 
    int str;
    
    const_string(byte* p) : constant(p) { 
	str = unpack2(p+1);
    }

    virtual int length() { return 3; }

    virtual int copy(generation_context& ctx, constant** pool, int* mapping) { 
	int new_str = relocate(str, ctx, pool, mapping);
	byte* dst = ctx.extend_constant_buffer(3);
	*dst++ = tag;
	pack2(dst, new_str);
	return ctx.constant_pool_count++;
    }
};

class const_class : public constant {
  public: 
    int name;

    const_class(byte* p) : constant(p) {
	name = unpack2(p+1);
    }

    virtual int length() { return 3; }

    virtual int copy(generation_context& ctx, constant** pool, int* mapping) { 
	int new_name = relocate(name, ctx, pool, mapping);
	byte* dst = ctx.extend_constant_buffer(3);
	*dst++ = tag;
	pack2(dst, new_name);
	return ctx.constant_pool_count++;
    }
};

class const_ref : public constant {
  public:
    int cls;
    int name_and_type;

    const_ref(byte* p) : constant(p) {
	cls = unpack2(p+1);
	name_and_type = unpack2(p+3);
    }

    virtual int length() { return 5; }

    virtual int copy(generation_context& ctx, constant** pool, int* mapping) { 
	int new_cls = relocate(cls, ctx, pool, mapping);
	int new_nt = relocate(name_and_type, ctx, pool, mapping);
	byte* dst = ctx.extend_constant_buffer(5);
	*dst++ = tag;
	dst = pack2(dst, new_cls);
	pack2(dst, new_nt);
	return ctx.constant_pool_count++;
    }
};


class const_name_and_type : public constant {
  public:
    int name;
    int desc;

    const_name_and_type(byte* p) : constant(p) {
	name = unpack2(p+1);
	desc = unpack2(p+3);
    }

    virtual int length() { return 5; }

    virtual int copy(generation_context& ctx, constant** pool, int* mapping) { 
	int new_name = relocate(name, ctx, pool, mapping);
	int new_desc = relocate(desc, ctx, pool, mapping);
	byte* dst = ctx.extend_constant_buffer(5);
	*dst++ = tag;
	dst = pack2(dst, new_name);
	pack2(dst, new_desc);
	return ctx.constant_pool_count++;
    }
};



class class_method { 
  public:
    class_method* next;
    
    utf_string const& name;
    utf_string const& desc;

    class_file* this_class;

    int         attr;
    int         max_stack, new_max_stack;
    int         max_locals, new_max_locals;
    int         code_length;
    int         new_code_length;
    int         n_parameters;
    byte*       param_type;

    byte*       code;
    byte*       new_code;

    byte*       exception_table;
    byte*       new_exception_table;

    bool        recursive;

    int         exception_table_length;
    int         new_exception_table_length;
    int         line_table_size;
    int         local_var_table_size;

    int         call_count;

    class_method(utf_string const& mth_name, utf_string const& mth_desc,
		 int mth_attr, class_file* cls); 

    void optimize(generation_context& ctx, bool inline_function);
};

#define CLASS_HASH_TABLE_SIZE 1013

class class_file { 
  public:
    class_file* next;
    class_file* collision_chain;
    static class_file* class_list;
    static class_file* hash_table[CLASS_HASH_TABLE_SIZE];

    char*       file_name;
    utf_string* class_name;
    utf_string* super_class_name;
    int         this_class;

    constant**  constant_pool;
    int         constant_pool_count;

    int*        constant_mapping;
    class_file* mapping_for_class;

    bool        archived;
    bool        final;

    byte*       buffer;
    byte*       constants;
    byte*       attributes;
    byte*       methods;

    class_method* method_list;

    static generation_context ctx;

    void optimize_file();
    static void optimize();

    class_method* find_method(utf_string const& name, utf_string const& desc);
	    
    static class_file* find(utf_string const& name);

    class_file(const char* file_name, int file_name_len, byte* fp, bool zip);
};

class relocation_record { 
  public:
    relocation_record* next;
    int                target_pc;
    int                inst_pc;
    int                size;
    int                offs;
    byte*              ptr;
    
    static relocation_record* free;

    static void add(relocation_record* &chain, int old_pc, int new_pc, 
		    byte* ptr, int size, int* relocation_table)
    {
	int offs;
	if (size == 2) { 
	    offs = (short)unpack2(ptr); 
	    if (offs <= 0) { 
		pack2(ptr, relocation_table[old_pc+offs] - new_pc);
		return;
	    } 
	} else { 
	    offs = unpack4(ptr); 
	    if (offs <= 0) { 
		pack4(ptr, relocation_table[old_pc+offs] - new_pc);
		return;
	    } 
	}
	relocation_record* rp = free;
	if (rp == NULL) { 
	    rp = new relocation_record;
	} else { 
	    free = rp->next;
	}
	relocation_record** rpp;
	old_pc += offs;
	for (rpp = &chain; 
	     *rpp != NULL && (*rpp)->target_pc < old_pc; 
	     rpp = &(*rpp)->next); 
	rp->next = *rpp;
	*rpp = rp;
	rp->target_pc = old_pc;
	rp->inst_pc = new_pc;
	rp->size = size;
	rp->ptr = ptr;
    }

    static bool check(relocation_record* &chain, int old_pc, int new_pc) 
    { 
	relocation_record* rp = chain;
	bool entry_point = false;
	while (rp != NULL && rp->target_pc <= old_pc) { 
	    assert (rp->target_pc == old_pc);
	    if (rp->size == 2) { 
		pack2(rp->ptr, new_pc - rp->inst_pc);
	    } else { 
		pack4(rp->ptr, new_pc - rp->inst_pc);
	    }
	    relocation_record* next = rp->next;
	    rp->next = free;
	    free = rp;
	    rp = next;
	    entry_point = true;
	}
	chain = rp;
	return entry_point;
    }
};

//
// Constants for extracting zip file
//

#define LOCAL_HDR_SIG     "\113\003\004"   /*  bytes, sans "P" (so unzip */
#define LREC_SIZE     26    /* lengths of local file headers, central */
#define CREC_SIZE     42    /*  directory headers, and the end-of-    */
#define ECREC_SIZE    18    /*  central-dir record, respectively      */
#define TOTAL_ENTRIES_CENTRAL_DIR  10
#define SIZE_CENTRAL_DIRECTORY     12
#define C_UNCOMPRESSED_SIZE        20
#define C_FILENAME_LENGTH          24
#define C_EXTRA_FIELD_LENGTH       26
#define C_RELATIVE_OFFSET_LOCAL_HEADER    38
#define L_FILENAME_LENGTH                 22
#define L_EXTRA_FIELD_LENGTH              24


#endif
